/*********************************************************************
 * popproxy version 1.3    24/2/97 by N.Oren                         *
 * used to convert long user names to unix user names and connect to *
 * the appropriate mail hub. 					     *
 *Changes from 1.3:Fixed a solaris bug            		     *
 *Changes from 1.11:Socket reads are now blocking	 	     *		
 *Changes from 1.1: A few small bugfixes, now looks more messy	     *
 *Changes from 1.03: Read now uses UNIX read and write instead of    *
 *                   streams (doesn't sound like an improvement      *
 *                   I know).                                        *
 *                   Works on Solaris                                *
 *                   Added UIDL command support                      *
 *Changes from 1.02:Ignoring spaces etc, but no luck, and added      *
 *                  a timeout on server side also                    *
 *Changes from 1.01:Fixed a small bug in the LIST command            *
 *Changes from 1.0: Now listens on port on all interfaces            *
 *Changes from 0.2: Now can use UNIX database to store hash table    *
 *Changes from 0.1: Can use either encrypted or unencrypted passwords*
 *********************************************************************/

#include "popproxy.h"

/***************************************************************
 makehash creates hash table in memory or on disk
***************************************************************/
void makehash(void)
{
#ifdef GDBM
   GDBM_FILE dbf;
   dbf=gdbm_open("popusers",0,GDBM_NEWDB,S_IRUSR|S_IWUSR,0);
   if (dbf==NULL)
     {
       perror("Unable to create database\n");
       exit(-1);
     };
#endif
   char *dummy=new char[1000]; /*hope its big enough*/
   char *name=new char[1000];  /*see above comment*/
   FILE *thefile=fopen(DEFFILE,"r");
   while (fgets(dummy,999,thefile))
     {
        if ((dummy[0]!='\n')&(dummy[0]!='#'))
          {
            char *first=strchr(dummy,':');
            strncpy(name,dummy,first-dummy);
            name[first-dummy]=0x0;
            first++;
#ifndef GDBM
            db.addent(name,first);
#endif
#ifdef GDBM
	    datum temp1;
            datum temp2;
            temp1.dptr=name;
            temp1.dsize=strlen(name)+1;
            temp2.dptr=first;
            temp2.dsize=strlen(first)+1;
            gdbm_store(dbf,temp1,temp2,GDBM_REPLACE);
#endif
          };
     };
   fflush(NULL);
   fclose(thefile);
   delete dummy;
   delete name;
#ifdef GDBM
   gdbm_close(dbf);
#endif
};


/****************************************************************
 getinfo slaps in the userdate into a popinfo structure, returning
 -1 on error
*****************************************************************/
int getinfo(char *username, popinfo *info)
{
  char *data=new char[1000];
#ifndef GDBM
  db.getdata(username,data);
#endif
#ifdef GDBM
  GDBM_FILE dbf=gdbm_open("popusers",0,GDBM_READER,0,0);
  datum temp1,temp2;
  temp1.dptr=username;
  temp1.dsize=strlen(username)+1;
  temp2=gdbm_fetch(dbf,temp1);
  if (temp2.dptr==NULL)
    { 
      delete data;
      gdbm_close(dbf);
      return -1;
    };
  strncpy(data,temp2.dptr,999);
  gdbm_close(dbf);
#endif
  if (data[0]==0x0)
    return -1;
  strncpy(info->fullname,username,99);
  strncpy(info->accntname,strtok(data,":"),9);
  strncpy(info->mailhub,strtok(NULL,":"),99);
  strncpy(info->pwd,strtok(NULL,"\n"),14);
  delete data;
  return 1;
};
/***********************************************************
 what to do if the alarm goes off
************************************************************/
void timeout(int signal)
{
    instream.~sockbuf();
  exit(-1);
};
  
/***********************************************************
 handle child processes
************************************************************/
void child_handle(int signal)
{
  wait(NULL);
};

/***********************************************************
 handle sighup
************************************************************/
void reload(int signal)
{
  makehash();
};

/***********************************************************
SIGTERM and SIGKILL are handled here as follows:
************************************************************/
void die(int signal)
{
/*  if (instream!=NULL)*/
    instream.~sockbuf();
  exit(-1);
};

/****************************************************************
Dying with grace
*****************************************************************/
void panic(void)
{
  mailhub.~sockbuf();
  instream.~sockbuf();
  exit(-1);
};

/****************************************************************
  This function parses and validates the username and password
  returns:
       0     quit
       1     Valid Username/password , info in popinfo
*****************************************************************/
int validate(popinfo &userinfo)
{
  char givenname[100];
  char givenpwd[100];
  char buffer[1000];
  char gotname=0;
  instream.writel("+OK Hiya....\n");
  fflush(NULL);
  while (1)
    {
      alarm(TIMEOUT);
      char whattodo[5];
      if (instream.readl(buffer)<1) /*this should be done*/
        {                            /*for all reads*/
          instream.~sockbuf();
          exit(-1);
        };
          
      alarm(0);
      if (strlen(buffer)>1)
        buffer[strlen(buffer)-2]=0x0;
      strncpy(whattodo,buffer,4);
      whattodo[4]=0x0;
      if (strcasecmp(whattodo,"QUIT")==0)
        {
           instream.writel("+OK bye...\r\n");
           return 0;
        };
      if (strcasecmp(whattodo,"USER")==0)
        {
          instream.writel("+OK password???\r\n");
          strncpy(givenname,&buffer[5],sizeof(givenname));
          gotname=1;
        }  
      else if (strcasecmp(whattodo,"PASS")==0)
        {
          switch(gotname)
            {
              case 0: instream.writel("-ERR need a username first\r\n");
                      break;
              case 1: switch(getinfo(givenname,&userinfo))
                        {
                          case -1: break;
                          case 1: strncpy(givenpwd,&buffer[5],sizeof(givenpwd));
#ifdef ENCRYPT
			          char salt[3];
                                  salt[0]=userinfo.pwd[0];
				  salt[1]=userinfo.pwd[1];
				  salt[2]=0x0;
                                  if (strcmp(userinfo.pwd,crypt(givenpwd,salt))==0)
                                    strncpy(userinfo.pwd,givenpwd,sizeof(givenpwd));
#endif
                                  if (strcmp(userinfo.pwd,givenpwd)==0)
                                    {
                                      alarm(0);
                                      syslog(LOG_INFO,"User %s retrieved his mail\n",givenname);
                                      return 1;
                                    };
                                  break;            
                        };
                      instream.writel("-ERR invalid username/password\r\n");
                      sleep(1);
                      gotname=0;
		      syslog(LOG_NOTICE,"User %s gave an incorrect username/password combination\n",givenname);	
                      break;
            };  
        }
      else
        {
          instream.writel("-ERR huh???\r\n");
          fflush(NULL);
          gotname=0;
        };
    };
  return 1;
};


/**********************************************************************
 The following function connects to the pop server, and
 sends all the validation data to it, once it is happy, it returns the
 file stream, and prints the last '+OK xx messages waiting' thing.
 NOTE:errorchecking should occur
***********************************************************************/
int connecttohost(popinfo userinfo)
{
  hostent *hp=gethostbyname(userinfo.mailhub);
  sockaddr_in sa;
  sa.sin_family=PF_INET;
  sa.sin_port=htons(HUBPORT);
  sa.sin_addr= *(in_addr *) hp->h_addr;
  int hubsock=socket(AF_INET,SOCK_STREAM,0);
  if (connect(hubsock,(sockaddr *) &sa,sizeof(sa))==-1)
    {
      syslog(LOG_ALERT,"Unable to connect to pop host!!\n");
      instream.~sockbuf();                  /*PANIC ERROR!!!*/
      exit(-1);
    };
  char buffer[1000]; /*should be long enough*/
  /*I'm not testing to see if they are right or wrong, I assume they are
    right, but I give a TIMOUT*5 timeout just in case*/
  alarm(TIMEOUT*5);
  read(hubsock,buffer,sizeof(buffer)-1);
  sprintf(buffer,"USER %s\r\n",userinfo.accntname);
  write(hubsock,buffer,strlen(buffer));
  read(hubsock,buffer,sizeof(buffer)-1);
  sprintf(buffer,"PASS %s\r\n",userinfo.pwd);
  write(hubsock,buffer,strlen(buffer));
  int length=read(hubsock,buffer,sizeof(buffer)-1);
  buffer[length]=0x0;
  instream.writel(buffer);
  alarm(0);
  return hubsock;
};

/**********************************************************************
handle the socket given to us by main, make it into a stream etc
etc, validate it, connect to host, and then handle data transfer
***********************************************************************/
int handle(int sock)
{
  instream.sockbuf(sock,1000);
  char buffer[1000];
  /*sockbuf mailhub; pop server we will connect to*/
  popinfo user;
  if (!validate(user))
    {
      fflush(NULL);
      instream.~sockbuf();
      return -1;
    };
  mailhub.sockbuf(connecttohost(user),1000);
  /*******
   OK, the LIST and RETR  and TOP commands are the only 3 which return more
   then one line, also,garbage from the client should be handled
   I'm also assuming that the hub is a good boy, and that
   the client is probably the guy who will cause us problems with timeouts
  ********/
  while(1)
    {  
      char waitfullstop=0;
      char quit=0;
      alarm(TIMEOUT);
      if (instream.readl(buffer)<1)  
        panic();
      alarm(0);
      if (strncasecmp(buffer,"QUIT",4)==0)
        quit=1;
      if (strncasecmp(buffer,"RETR",4)==0||strncasecmp(buffer,"TOP ",4)==0)
        waitfullstop=1;
      /* List can accept 1 or 0 params, and returns diffrent things */
      if ((atoi(&buffer[5])==0)&&(strncasecmp(buffer,"LIST",4)==0)) 
        waitfullstop=1;
      if ((atoi(&buffer[5])==0)&&(strncasecmp(buffer,"UIDL",4)==0)) 
        waitfullstop=1;
      /*Added 25/2/97 To check for lines containing only spaces*/
      while (isspace(buffer[0]))
          strcpy(buffer,&buffer[1]);
      if (buffer[0]!=0x0)
        { 
          if (mailhub.writel(buffer)==-1)
            panic();
          fflush(NULL);
          switch(waitfullstop)
            {
              case 0:alarm(TIMEOUT);
                     if (mailhub.readl(buffer)<1)
                       panic();
                     alarm(0);
                     if (instream.writel(buffer)==-1)
                       panic();
                     break;
              case 1:char stop=0;                /*THIS NEEDS TO BE*/
                     alarm(TIMEOUT);             /*TIDIED UP WITH A*/
                     if (mailhub.readl(buffer)<1)/*DO WHILE LOOP*/
                       panic();
                     alarm(0);                   /*RATHER THEN USING*/
                     if(instream.writel(buffer)==-1)    /*A WHILE LOOP!!!!!*/
                       panic();
                     if (strncasecmp(buffer,"-ERR",3)==0)
                       stop=1;
                     while (!stop)
                       { 
                         if (buffer[0]=='.' && buffer[1]=='\r')
                           stop=1;
                         else
                           {
                             alarm(TIMEOUT);
                             if (mailhub.readl(buffer)<1)
                               panic();
                             alarm(0);
                             if (instream.writel(buffer)==-1)
                               panic();
                           };
                       };
                    break;
            };
        }; /*end if buffer!=0x0*/ 
      if (quit)
        {
          mailhub.~sockbuf();
          instream.~sockbuf();
          exit(1); /*rather then return*/
        };
    };
};


/*********************************************************************
MAIN LOOP, become daemon, set up basic socket handling, and signal
handling. get hash table ready
*********************************************************************/
int main(void)
{

/* make hash table:*/

makehash();

/* Open logging */

openlog("POPPROXY",LOG_PID,LOG_DAEMON);

/*listen on sockets*/

  sockaddr_in sa;
  sockaddr other; /*addr of incoming socket*/
  int othersize; /*size of other*/
  int incoming=socket(AF_INET,SOCK_STREAM,0); /*incoming socket*/
  sa.sin_family=PF_INET;
  sa.sin_port=htons(LISTENON);
#ifdef LINUX
  sa.sin_addr.s_addr=(__u32) INADDR_ANY; /*there is probably a neater way*/
#endif
#ifdef SPARC
  sa.sin_addr.s_addr=INADDR_ANY;
#endif

  while (bind(incoming,(sockaddr *) &sa,sizeof(struct sockaddr_in))==-1);
      /* loop until socket is bound,this bit is a serious mess*/
  fcntl(incoming,F_SETFL,~O_NONBLOCK); /*make sure it blocks*/
  listen(incoming,5);

/* Become a daemon */

  switch(fork())
    {
      case 0:fclose(stdin);
             fclose(stdout);
             fclose(stderr);
             setsid();
             break;
      default: exit(0);
    };

/* handle signals: */

  signal(SIGALRM,timeout);
  signal(SIGCHLD,child_handle);
  signal(SIGHUP,reload);
  signal(SIGTERM,die);
  signal(SIGKILL,die);
  signal(SIGINT,SIG_IGN);
  signal(SIGPIPE,SIG_IGN); /*Needed*/

  while(1)
    {
      int sock=accept(incoming,&other,&othersize);
      if (sock!=-1)
        {
          switch(fork())
            {
              case 0: close(incoming);
                      handle(sock);
                      exit(0);
              default:close(sock);
                      signal(SIGCHLD,child_handle);
                      break;
            };
        };
    };
};

